# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""
SQL backend for syncstorage.

This module implements an SQL storage plugin for synctorage.  In the simplest
use case it consists of three database tables:

  collections:  the names and ids of any custom collections
  collection_timestamps:  the last-edit timestamps for each collection
  bso:  the individual BSO items stored in each collection

For efficiency when dealing with large datasets, the plugin also supports
sharding of the BSO items into multiple tables named "bso0" through "bsoN".
This behaviour is off by default; pass shard=True to enable it.
"""

import functools
import threading
from collections import defaultdict

from sqlalchemy.exc import IntegrityError

from pyramid.threadlocal import get_current_registry

from syncstorage.bso import BSO
from syncstorage.util import get_timestamp, from_timestamp
from syncstorage.storage import (SyncStorage,
                                 LockContextManager,
                                 ConflictError,
                                 CollectionNotFoundError,
                                 ItemNotFoundError)

from syncstorage.storage.sql.dbconnect import DBConnector, MAX_TTL

# For efficiency, it's possible to use fixed pre-determined IDs for
# common collection names.  This is the canonical list of such names.
# Non-standard collections will be allocated IDs starting from the
# highest ID in this collection.
STANDARD_COLLECTIONS = {1: "clients", 2: "crypto", 3: "forms", 4: "history",
                        5: "keys", 6: "meta", 7: "bookmarks", 8: "prefs",
                        9: "tabs", 10: "passwords", 11: "addons"}

FIRST_CUSTOM_COLLECTION_ID = 100

MAX_COLLECTIONS_CACHE_SIZE = 1000


def with_session(func):
    """Method decorator to magic a "session" object into existence.

    The SQLStorage backend uses a session object to manage database connections
    and related internal state.  This is not exposed in the public API, but
    is instead managed through a threadlocal storage object.

    This method decorator obtains the active session for the current thread
    (creating one if it doesn't exist) and passes it as the first argument
    to the underlying method.
    """
    @functools.wraps(func)
    def with_session_wrapper(self, *args, **kwds):
        # If the first argument is already a session object, just use that.
        if args and isinstance(args[0], SQLStorageSession):
            return func(self, *args, **kwds)
        # Otherwise, magic one into existence using threadlocals.
        try:
            session = self._tldata.session
        except AttributeError:
            session = SQLStorageSession(self)
        with session:
            return func(self, session, *args, **kwds)
    return with_session_wrapper


class SQLStorage(SyncStorage):
    """Storage plugin implemented using an SQL database.

    This class implements the storage plugin API using SQLAlchemy.  You
    must specify the SQLAlchemy database URI string to connect to, and
    can customize behaviour with the following keyword arguments:

        * standard_collections:  use fixed pre-determined ids for common
                                 collection names
        * create_tables:         create the database tables if they don't
                                 exist at startup
        * shard/shardsize:       enable sharding of the BSO table

    """

    def __init__(self, sqluri, standard_collections=False, **dbkwds):

        self.logger = get_current_registry()["metlog"]

        self.sqluri = sqluri
        self.dbconnector = DBConnector(sqluri, **dbkwds)

        # There doesn't seem to be a reliable cross-database way to set the
        # initial value of an autoincrement column.  Fake it by inserting
        # a row into the table at the desired start id.
        self.standard_collections = standard_collections
        if self.standard_collections and dbkwds.get("create_tables", False):
            zeroth_id = FIRST_CUSTOM_COLLECTION_ID - 1
            with self.dbconnector.connect() as connection:
                params = {"collectionid": zeroth_id, "name": ""}
                try:
                    connection.query("INSERT_COLLECTION", params)
                except IntegrityError:
                    pass

        # A local in-memory cache for the name => collectionid mapping.
        self._collections_by_name = {}
        self._collections_by_id = {}
        if self.standard_collections:
            for id, name in STANDARD_COLLECTIONS:
                self._collections_by_name[name] = id
                self._collections_by_id[id] = name

        # A thread-local to track active sessions.
        self._tldata = threading.local()

    #
    # APIs for collection-level locking.
    #
    # For SQLite we depend on its database-level locking.  Read locks begin
    # a transaction and select something from the database, which creates
    # a SHARED lock.  Write locks begin an immediate transaction, which
    # creates a RESERVED lock on the database.
    #
    # For MySQL (and other unsupported databases) we do explicit locking
    # on the matching row in the user_collections table.  Read locks do
    # SELECT ... LOCK IN SHARE MODE and write locks do SELECT ... FOR UPDATE.
    #

    @with_session
    def lock_for_read(self, session, userid, collection):
        """Acquire a shared read lock on the named collection."""
        collectionid = self._get_collection_id(session, collection)
        if (userid, collectionid) in session.locked_collections:
            raise RuntimeError("Collection already locked")
        session.query("BEGIN_TRANSACTION_READ")
        params = {"userid": userid, "collectionid": collectionid}
        ts = session.query_scalar("LOCK_COLLECTION_READ", params)
        if ts is None:
            raise CollectionNotFoundError
            session.cache[(userid, collectionid)].last_modified = ts
        # If that all completes successfully, re-enter the session context
        # so that it sticks around after this function exists.  A future
        # call to unlock() will exit the session.
        session.begin()
        session.locked_collections.add((userid, collection))
        return LockContextManager(self, userid, collection)

    @with_session
    def lock_for_write(self, session, userid, collection):
        """Acquire an exclusive write lock on the named collection."""
        # XXX TODO: need to fail out with ConflictError if lock already held.
        collectionid = self._get_collection_id(session, collection, create=1)
        if (userid, collectionid) in session.locked_collections:
            raise RuntimeError("Collection already locked")
        session.query("BEGIN_TRANSACTION_WRITE")
        params = {"userid": userid, "collectionid": collectionid}
        ts = session.query_scalar("LOCK_COLLECTION_WRITE", params)
        if ts is not None:
            session.cache[(userid, collectionid)].last_modified = ts
        # If that all completes successfully, re-enter the session context
        # so that it sticks around after this function exists.  A future
        # call to unlock() will exit the session.
        session.begin()
        session.locked_collections.add((userid, collection))
        return LockContextManager(self, userid, collection)

    @with_session
    def unlock(self, session, userid, collection):
        """Release any locks acquired on the given collection.

        This method releases collection locks by committing the active
        transactions.  This means that no locks will actually be released
        until *all* locks are released, which should not be a problem given
        the intended usecase.
        """
        collectionid = self._get_collection_id(session, collection, create=1)
        if (userid, collectionid) not in session.locked_collections:
            raise RuntimeError("Collection is not locked")
        session.locked_collections.remove((userid, collection))
        session.commit()

    #
    # APIs to operate on the entire storage.
    #

    @with_session
    def get_storage_timestamp(self, session, userid):
        """Returns the last-modified timestamp for the entire storage."""
        return session.query_scalar("STORAGE_TIMESTAMP", params={
            "userid": userid,
        }, default=0)

    @with_session
    def get_collection_timestamps(self, session, userid):
        """Returns the collection timestamps for a user."""
        res = session.query_fetchall("COLLECTIONS_TIMESTAMPS", {
            "userid": userid,
        })
        return self._map_collection_names(session, res)

    @with_session
    def get_collection_counts(self, session, userid):
        """Returns the collection counts."""
        res = session.query_fetchall("COLLECTIONS_COUNTS", {
            "userid": userid,
            "ttl": int(from_timestamp(session.timestamp))
        })
        return self._map_collection_names(session, res)

    @with_session
    def get_collection_sizes(self, session, userid):
        """Returns the total size for each collection."""
        res = session.query_fetchall("COLLECTIONS_SIZES", {
            "userid": userid,
            "ttl": int(from_timestamp(session.timestamp)),
        })
        return self._map_collection_names(session, res)

    @with_session
    def get_total_size(self, session, userid, recalculate=False):
        """Returns the total size a user's stored data."""
        return session.query_scalar("STORAGE_SIZE", {
            "userid": userid,
            "ttl": int(from_timestamp(session.timestamp)),
        }, default=0)

    @with_session
    def delete_storage(self, session, userid):
        """Removes all data for the user."""
        session.query("DELETE_ALL_BSOS", {
            "userid": userid,
        })
        session.query("DELETE_ALL_COLLECTIONS", {
            "userid": userid,
        })

    #
    # APIs to operate on an individual collection
    #

    @with_session
    def get_collection_timestamp(self, session, userid, collection):
        """Returns the last-modified timestamp of a collection."""
        collectionid = self._get_collection_id(session, collection)
        # The last-modified timestamp may be cached on the session.
        cached_timestamp = session.cache[(userid, collectionid)].last_modified
        if cached_timestamp is not None:
            return cached_timestamp
        # Otherwise we need to look it up in the database.
        timestamp = session.query_scalar("COLLECTION_TIMESTAMP", {
            "userid": userid,
            "collectionid": collectionid,
        })
        if timestamp is None:
            raise CollectionNotFoundError
        return timestamp

    @with_session
    def get_items(self, session, userid, collection, **params):
        """Returns items from a collection."""
        return self._find_items(session, userid, collection, **params)

    @with_session
    def get_item_ids(self, session, userid, collection, **params):
        """Returns item ids from a collection."""
        params["fields"] = ["id"]
        items = self._find_items(session, userid, collection, **params)
        return [item["id"] for item in items]

    def _find_items(self, session, userid, collection, **params):
        """Find items matching the given search parameters."""
        params["userid"] = userid
        params["collectionid"] = self._get_collection_id(session, collection)
        if "ttl" not in params:
            params["ttl"] = int(from_timestamp(session.timestamp))
        rows = session.query_fetchall("FIND_ITEMS", params)
        items = [self._row_to_bso(row) for row in rows]
        # If the query returned no results, we don't know whether that's
        # because it's empty or because it doesn't exist.  Read the collection
        # timestamp and let it raise CollectionNotFoundError if necessary.
        if not items:
            self.get_collection_timestamp(session, userid, collection)
        return items

    def _row_to_bso(self, row):
        """Convert a database table row into a BSO object."""
        item = dict(row)
        for key in ("userid", "collection", "payload_size", "ttl",):
            item.pop(key, None)
        return BSO(item)

    @with_session
    def set_items(self, session, userid, collection, items):
        """Creates or updates multiple items in a collection."""
        collectionid = self._get_collection_id(session, collection, create=1)
        for data in items:
            id = data["id"]
            self._prepare_item_data(session, userid, collectionid, id, data)
        session.insert_or_update("bso", items)
        self._touch_collection(session, userid, collectionid)
        return session.timestamp

    @with_session
    def delete_collection(self, session, userid, collection):
        """Deletes an entire collection."""
        collectionid = self._get_collection_id(session, collection)
        count = session.query("DELETE_COLLECTION_ITEMS", {
            "userid": userid,
            "collectionid": collectionid,
        })
        count += session.query("DELETE_COLLECTION", {
            "userid": userid,
            "collectionid": collectionid,
        })
        if count == 0:
            raise CollectionNotFoundError
        return self.get_storage_timestamp(userid)

    @with_session
    def delete_items(self, session, userid, collection, items):
        """Deletes multiple items from a collection."""
        collectionid = self._get_collection_id(session, collection)
        session.query("DELETE_ITEMS", {
            "userid": userid,
            "collectionid": collectionid,
            "ids": items,
        })
        self._touch_collection(session, userid, collectionid)
        return session.timestamp

    def _touch_collection(self, session, userid, collectionid):
        """Update the modified time of the given collection."""
        params = {
            "userid": userid,
            "collectionid": collectionid,
            "modified": session.timestamp,
        }
        # The common case will be an UPDATE, so try that first.
        # If it doesn't update any rows then do an INSERT.
        # XXX TODO: we should refuse to move the timestamp backwards.
        rowcount = session.query("TOUCH_COLLECTION", params)
        if rowcount != 1:
            try:
                session.query("INIT_COLLECTION", params)
            except IntegrityError:
                # Someone else inserted it at the same time.
                pass

    #
    # Items APIs
    #

    @with_session
    def get_item_timestamp(self, session, userid, collection, item):
        """Returns the last-modified timestamp for the named item."""
        collectionid = self._get_collection_id(session, collection)
        timestamp = session.query_scalar("ITEM_TIMESTAMP", {
            "userid": userid,
            "collectionid": collectionid,
            "item": item,
            "ttl": int(from_timestamp(session.timestamp)),
        })
        if timestamp is None:
            raise ItemNotFoundError
        return timestamp

    @with_session
    def get_item(self, session, userid, collection, item):
        """Returns one item from a collection."""
        collectionid = self._get_collection_id(session, collection)
        row = session.query_fetchone("ITEM_DETAILS", {
            "userid": userid,
            "collectionid": collectionid,
            "item": item,
            "ttl": int(from_timestamp(session.timestamp)),
        })
        if row is None:
            raise ItemNotFoundError
        return self._row_to_bso(row)

    @with_session
    def set_item(self, session, userid, collection, item, data):
        """Creates or updates a single item in a collection."""
        collectionid = self._get_collection_id(session, collection, create=1)
        self._prepare_item_data(session, userid, collectionid, item, data)
        # XXX TODO: need to know if created or updated
        try:
            self.get_item_timestamp(userid, collection, item)
            created = False
        except ItemNotFoundError:
            created = True
        session.insert_or_update("bso", [data])
        self._touch_collection(session, userid, collectionid)
        return {
            "created": created,
            "modified": session.timestamp,
        }

    def _prepare_item_data(self, session, userid, collectionid, item, data):
        """Fill in and normalize fields in the given item data."""
        data["userid"] = userid
        data["collection"] = collectionid
        data["id"] = item
        # If a payload is provided, make sure to update dependant fields.
        if "payload" in data:
            data["modified"] = session.timestamp
            data["payload_size"] = len(data["payload"])
        # If provided, ttl will be an offset in seconds.
        # Add it to the current timestamp to get an absolute time.
        if "ttl" not in data:
            data["ttl"] = MAX_TTL
        else:
            data["ttl"] += int(from_timestamp(session.timestamp))
        return data

    @with_session
    def delete_item(self, session, userid, collection, item):
        """Deletes a single item from a collection."""
        collectionid = self._get_collection_id(session, collection)
        rowcount = session.query("DELETE_ITEM", {
            "userid": userid,
            "collectionid": collectionid,
            "item": item,
            "ttl": int(from_timestamp(session.timestamp)),
        })
        if rowcount == 0:
            raise ItemNotFoundError
        self._touch_collection(session, userid, collection)
        return session.timestamp

    #
    # Private methods for manipulating collections.
    #

    def _get_collection_id(self, session, collection, create=False):
        """Returns a collection id, given the name.

        If the named collection does not exist then CollectionNotFoundError
        will be raised.  To automatically create collections on demand, pass
        create=True.
        """
        # Grab it from the cache if we can.
        try:
            return self._collections_by_name[collection]
        except KeyError:
            pass

        # Try to look it up in the database.
        collectionid = session.query_scalar("COLLECTION_ID", {
            "name": collection
        })
        if collectionid is None:
            # Shall we auto-create it?
            if not create:
                raise CollectionNotFoundError
            # Insert it into the database.  This might raise a conflict
            # if it was inserted concurrently by someone else.
            try:
                session.query("INSERT_COLLECTION", {
                    "collectionid": None,
                    "name": collection,
                })
            except IntegrityError:
                pass
            # Read the id that was created concurrently.
            collectionid = self._get_collection_id(session, collection)

        # Sanity-check that we"re not trampling standard collection ids.
        if self.standard_collections:
            assert collectionid >= FIRST_CUSTOM_COLLECTION_ID

        self._cache_collection_id(collectionid, collection)
        return collectionid

    def _get_collection_name(self, session, collectionid):
        """Returns a collection name, given the id.

        If the collection id does not exist then CollectionNotFoundError
        will be raised.
        """
        try:
            return self._collections_by_id[collectionid]
        except KeyError:
            pass

        collection = session.query_scalar("COLLECTION_NAME", {
            "collectionid": collectionid,
        })
        if collection is None:
            raise CollectionNotFoundError
        self._cache_collection_id(collectionid, collection)
        return collection

    def _load_collection_names(self, session, collection_ids):
        """Load any uncached names for the given collection ids.

        If you have a list of collection ids and you want all their names,
        use this method to prime the internal name cache.  Otherwise you"ll
        cause _get_collection_name() to do a separate database query for
        each collection id, which is very inefficient.

        Since it may not be possible to cache the names for all the request
        collections, this method also returns a mapping from ids to names.
        """
        names = {}
        uncached_ids = []
        # Extract as many names as possible from the cache, and
        # build a list of any ids whose names are not cached.
        for id in collection_ids:
            try:
                names[id] = self._collections_by_name[id]
            except KeyError:
                uncached_ids.append(id)
        # Use a single query to fetch the names for all uncached collections.
        if uncached_ids:
            uncached_names = session.query_fetchall("COLLECTION_NAMES", {
                "ids": uncached_ids,
            })
            for id, name in uncached_names:
                names[id] = name
                self._cache_collection_id(id, name)
        return names

    def _map_collection_names(self, session, values):
        """Helper to create a map of collection names to values.

        Given a sequence of (collectionid, value) pairs, this method will
        return a mapping from collection names to their corresponding value.
        """
        values = list(values)
        collection_ids = [collectionid for collectionid, value in values]
        names = self._load_collection_names(session, collection_ids)
        return dict([(names[id], value) for id, value in values])

    def _cache_collection_id(self, collectionid, collection):
        """Cache the given collection (id, name) pair for fast lookup."""
        if len(self._collections_by_name) > MAX_COLLECTIONS_CACHE_SIZE:
            msg = "More than %d collections have been created, "\
                  "refusing to cache them all"
            self.logger.warn(msg % (MAX_COLLECTIONS_CACHE_SIZE,))
        else:
            self._collections_by_name[collection] = collectionid
            self._collections_by_id[collectionid] = collection


class SQLStorageSession(object):
    """Object representing a data access session.

    The SQLStorageSession object is used to encapsulate a database connection
    and transaction, along with some metadata about the snapshot of data
    begin accessed.  For example:

        * the "current time" on the server during the snapshot
        * the last-modified times of collections
        * the set of currently-locked collections
    """

    def __init__(self, storage):
        self.storage = storage
        self.connection = storage.dbconnector.connect()
        self.timestamp = get_timestamp()
        self.cache = defaultdict(SQLCachedCollectionData)
        self.locked_collections = set()
        self._nesting_level = 0

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.commit()
        else:
            self.rollback()

    def insert_or_update(self, table, items):
        """Do a bulk insert/update of the given items."""
        return self.connection.insert_or_update(table, items)

    def query(self, query, params):
        """Execute a database query, returning the rowcount."""
        return self.connection.query(query, params)

    def query_scalar(self, query, params, default=None):
        """Execute a database query, returning a single scalar value."""
        return self.connection.query_scalar(query, params, default)

    def query_fetchone(self, query, params):
        """Execute a database query, returning the first result."""
        return self.connection.query_fetchone(query, params)

    def query_fetchall(self, query, params):
        """Execute a database query, returning iterator over the results."""
        return self.connection.query_fetchall(query, params)

    def begin(self):
        """Enter the context of this session.

        Calling this method marks the session as active for the current thread.
        It can be called multiple times, each requiring a corresponding call
        to either commit() or rollback().
        """
        if self._nesting_level == 0:
            assert not hasattr(self.storage._tldata, "session")
            self.storage._tldata.session = self
        self._nesting_level += 1

    def commit(self):
        """Successfully exit the context of this session.

        Once each entered context has been exited, this method will commit
        the underlying database transaction and close the connection.
        """
        self._nesting_level -= 1
        assert self._nesting_level >= 0
        if self._nesting_level == 0:
            del self.storage._tldata.session
            self.connection.commit()
            if self.locked_collections:
                msg = "You must unlock all collections before ending a session"
                raise RuntimeError(msg)

    def rollback(self):
        """Unsuccessfully exit the context of this session.

        Once each entered context has been exited, this method will rollback
        the underlying database transaction and close the connection.
        """
        self._nesting_level -= 1
        assert self._nesting_level >= 0
        if self._nesting_level == 0:
            del self.storage._tldata.session
            self.connection.rollback()
            if self.locked_collections:
                msg = "You must unlock all collections before ending a session"
                raise RuntimeError(msg)


class SQLCachedCollectionData(object):
    """Object for storing cached information about a collection.

    The SQLStorageSession object maintains a small cache of data that has
    already been looked up during that session.  Currently this includes only
    the last-modified timestamp of any collections locked by that session.
    """
    def __init__(self):
        self.last_modified = None
