# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Low-level SQL backend for syncstorage.

This module implements a thin data access layer on top of an SQL database,
providing the primitive operations on which to build a full SyncStorage
backend.  It provides three database tables:

  collections:  the names and ids of all collections in the store
  collection_timestamps:  the per-user timestamps for each collection
  bso:  the individual BSO items stored in each collection

For efficiency when dealing with large datasets, this module also supports
sharding of the BSO items into multiple tables named "bso0" through "bsoN".
This behaviour is off by default; pass shard=True to enable it.
"""

import urlparse
import traceback
from collections import defaultdict

from pyramid.threadlocal import get_current_registry

from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool
from sqlalchemy.sql import text as sqltext, insert, update
from sqlalchemy.exc import (DBAPIError, OperationalError,
                            TimeoutError, IntegrityError)
from sqlalchemy import (Integer, String, Text, BigInteger,
                        MetaData, Column, Table)

from mozsvc.exceptions import BackendError

from syncstorage.storage.sql import queries_generic, queries_sqlite

MAX_TTL = 2100000000

metadata = MetaData()


# Table mapping collection_name => collection_id.
#
# This table holds the names and corresponding ids of the collections in
# use on the storage node.  The collection id space is global, since we
# expect most users to have the same small, static set of collection names.

collections = Table("collections", metadata,
    Column("collectionid", Integer, primary_key=True, nullable=False,
                           autoincrement=True),
    Column("name", String(32), nullable=False, unique=True)
)


# Table mapping (user_id, collection_id) => collection-level metadata.
#
# This table holds collection-level metadata on a per-user basis.  Currently
# the only such metadata is the last-modified time of the collection.

user_collections = Table("user_collections", metadata,
    Column("userid", Integer, primary_key=True, nullable=False),
    Column("collection", Integer, primary_key=True, nullable=False),
    Column("last_modified", BigInteger, nullable=False)
)


# Column definitions for BSO storage table/tables.
#
# This list class defines the columns used for storage of BSO records.
# It is used to create either sharded or non-shareded BSO storage tables,
# depending on the run-time settings of the application.

def _get_bso_columns():
    return (
      Column("id", String(64), primary_key=True, autoincrement=False),
      Column("userid", Integer, primary_key=True, nullable=False),
      Column("collection", Integer, primary_key=True, nullable=False),
      Column("sortindex", Integer),
      Column("modified", BigInteger),
      Column("payload", Text, nullable=False, default=""),
      Column("payload_size", Integer, nullable=False, default=0),
      Column("ttl", Integer, default=MAX_TTL),
      # XXX TODO add index declarations
    )


#  If the storage controller is not doing sharding based on userid,
#  then it will use the single "bso" table below for BSO storage.

bso = Table("bso", metadata, *_get_bso_columns())

#  If the storage controller is doing sharding based on userid,
#  then it will use the below functions to select a table from "bso0"
#  to "bsoN" for each userid.

BSO_SHARDS = {}


def get_bso_table(index):
    """Get the Table object for table bso<N>."""
    bso = BSO_SHARDS.get(index)
    if bso is None:
        bso = Table("bso%d" % (index,), metadata, *_get_bso_columns())
        BSO_SHARDS[index] = bso
    return bso


class DBConnector(object):
    """Database connector class for SQL access layer.

    This class, along with its companion class DBConnection, provide the
    layer through which to access the SQL database.  It is a thin layer
    on top of the SQLAlchemy engine/connection machinery, with the following
    additional features:

        * transparent sharding of BSO storage tables
        * use pre-defined queries rather than inline construction of SQL
        * accessor methods that automatically clean up database resources
        * automatic retry of connections that are invalidated by the server

    """

    def __init__(self, sqluri, create_tables=False, pool_size=100,
                 no_pool=False, pool_recycle=60, reset_on_return=True,
                 pool_max_overflow=10, pool_timeout=30,
                 shard=False, shardsize=100, **kwds):

        self.logger = get_current_registry()["metlog"]

        self.sqluri = sqluri
        self.driver = urlparse.urlparse(sqluri).scheme.lower()
        if "mysql" in self.driver:
            self.driver = "mysql"

        if self.driver not in ("mysql", "sqlite"):
            msg = "Only MySQL and SQLite databases are officially supported"
            self.logger.warn(msg)

        self.shard = shard
        self.shardsize = shardsize

        # Create the SQLAlchemy engine, using the given parameters for
        # connection pooling.  Pooling doesn't work properly for sqlite so
        # it's disabled for that driver regardless of the value of no_pool.
        sqlkw = {}
        sqlkw["logging_name"] = "syncstorage"
        if no_pool or self.driver == 'sqlite':
            sqlkw["poolclass"] = NullPool
        else:
            sqlkw["pool_size"] = int(pool_size),
            sqlkw["pool_recycle"] = int(pool_recycle)
            sqlkw["pool_timeout"] = int(pool_timeout)
            sqlkw["max_overflow"] = int(pool_max_overflow)
            sqlkw["reset_on_return"] = reset_on_return
        self.engine = create_engine(sqluri, **sqlkw)

        # Create the tables if necessary.
        if create_tables:
            collections.create(self.engine, checkfirst=True)
            user_collections.create(self.engine, checkfirst=True)
            if not self.shard:
                bso.create(self.engine, checkfirst=True)
            else:
                for idx in xrange(self.shardsize):
                    bsoN = get_bso_table(idx)
                    bsoN.create(self.engine, checkfirst=True)

        # Load the pre-built queries to use with this database backend.
        # Currently we have a generic set of queries, and some queries specific
        # to SQLite.  We may add more backend-specific queries in future.
        self._prebuilt_queries = {}
        query_modules = [queries_generic]
        if self.driver == "sqlite":
            query_modules.append(queries_sqlite)
        for queries in query_modules:
            for nm in dir(queries):
                if nm.isupper():
                    self._prebuilt_queries[nm] = getattr(queries, nm)

    def connect(self, *args, **kwds):
        """Create a new DBConnection object from this connector."""
        return DBConnection(self)

    def get_query(self, name, params):
        """Get the named pre-built query.

        This method returns an SQLAlchemy query object for the named query,
        after performing some sharding based on the given parameters.
        """
        # Get the pre-built query with that name.
        # It might be None, a string query, or a callable returning the query.
        try:
            query = self._prebuilt_queries[name]
        except KeyError:
            raise KeyError("No query named %r" % (name,))
        # If it's None then just returning it, indicating a no-op.
        if query is None:
            return None
        # If it's a callable, call it with the sharded bso table.
        if callable(query):
            bso = self.get_bso_table(params.get("userid"))
            return query(bso, params)
        # If it's a string, do some interpolation and return it as sqltext.
        # XXX TODO: we could pre-parse these queries at load time to look for
        # string interpolation variables, saving some time on each call.
        assert isinstance(query, basestring)
        qvars = {}
        if "%(bso)s" in query:
            qvars["bso"] = self.get_bso_table(params["userid"])
        if "%(ids)s" in query:
            bindparams = []
            for i, id in enumerate(params["ids"]):
                params["id%d" % (i,)] = id
                bindparams.append(":id%d" % (i,))
            qvars["ids"] = "(" + ",".join(bindparams) + ")"
        if qvars:
            query = query % qvars
        return sqltext(query)

    def get_bso_table(self, userid):
        """Get the BSO table object for the given userid."""
        if not self.shard or userid is None:
            return bso
        return get_bso_table(userid % self.shardsize)


class DBConnection(object):
    """Database connection class for SQL access layer.

    This class provides a light abstraction around SQLAlchemy's Connection
    object.  It offers automatic retry of invalidated connections and some
    higher-level utility methods for running pre-written named queries.

    DBConnection classes always operate within a single, implicit database
    transaction.  The transaction is opened the first time a query is
    executed and is closed by calling either the commit() or rollback()
    method.
    """

    def __init__(self, connector):
        self._connector = connector
        self._connection = None
        self._transaction = None
        self.logger = connector.logger

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.commit()
        else:
            self.rollback()

    def commit(self):
        """Commit the active transaction and close the connection."""
        try:
            if self._transaction is not None:
                self._transaction.commit()
                self._transaction = None
        finally:
            if self._connection is not None:
                self._connection.close()
                self._connection = None

    def rollback(self):
        """Abort the active transaction and close the connection."""
        try:
            if self._transaction is not None:
                self._transaction.rollback()
                self._transaction = None
        finally:
            if self._connection is not None:
                self._connection.close()
                self._connection = None

    def execute(self, query, params={}):
        """Execute a database query, with retry and exception-catching logic.

        This method executes the given query against the database, lazily
        establishing an actual live connection as required.  It catches
        operational database errors and normalizes them into a BackendError
        exception.
        """
        try:
            # If there is no active connection, create a fresh one.
            # This will affect the controll flow below.
            connection = self._connection
            session_was_active = True
            if connection is None:
                connection = self._connector.engine.connect()
                transaction = connection.begin()
                session_was_active = False
            # It's possible for the backend to raise a "connection invalided"
            # error if e.g. the server timed out the connection we got from
            # the pool.  It's safe to retry with a new connection, but only
            # if the failed connection was never successfully used.
            try:
                return connection.execute(query, **params)
            except DBAPIError, exc:
                if not exc.connection_invalidated:
                    raise
                if session_was_active:
                    raise
                # The connection is dead, no need to close it here
                # before opening a fresh one.
                connection = self._connector.engine.connect()
                transaction = connection.begin()
                return connection.execute(query, **params)
        except (OperationalError, TimeoutError), exc:
            # Log the error, then normaize it into a BackendError instance.
            # Note that this will not not logic errors such as IntegrityError,
            # only unexpected operational errors from the database.
            err = traceback.format_exc()
            self.logger.error(err)
            raise BackendError(str(exc))
        finally:
            # Now that the underlying connection has been used, remember it
            # so that all subsequent queries are part of the same transaction.
            if not session_was_active:
                self._connection = connection
                self._transaction = transaction

    def query(self, query, params={}):
        """Execute a database query, returning the rowcount."""
        query = self._connector.get_query(query, params)
        res = self.execute(query, params)
        try:
            return res.rowcount
        finally:
            res.close()

    def query_scalar(self, query, params={}, default=None):
        """Execute a named query, returning a single scalar value."""
        query = self._connector.get_query(query, params)
        res = self.execute(query, params)
        try:
            row = res.fetchone()
            if row is None or row[0] is None:
                return default
            return row[0]
        finally:
            res.close()

    def query_fetchone(self, query, params={}):
        """Execute a named query, returning the first result row."""
        query = self._connector.get_query(query, params)
        res = self.execute(query, params)
        try:
            return res.fetchone()
        finally:
            res.close()

    def query_fetchall(self, query, params={}):
        """Execute a named query, returning iterator over the results."""
        query = self._connector.get_query(query, params)
        res = self.execute(query, params)
        try:
            for row in res:
                yield row
        finally:
            res.close()

    def insert_or_update(self, table, items):
        """Perform an efficient bulk "upsert" of the given items.

        Given the name of a table and a list of data dicts to insert or update,
        this method performs the "upsert" in the most efficient way.  It's
        a separate method because the precise details of the operation depend
        on the database driver in use.

        For generic database backends, the best we can do is try each insert,
        catch any IntegrityErrors and retry as an update.  For MySQL however
        we can use the "ON DUPLICATE KEY UPDATE" syntax to do the operation
        in a single query.
        """
        # Inserting zero items is strange, but allowed.
        if not items:
            return
        # Find the table object into which we're inserting.
        # To work properly with sharding, all items must have same userid
        # so that we can select a single BSO table.
        userid = items[0].get("userid")
        if table == "bso":
            table = self._connector.get_bso_table(userid)
        else:
            table = metadata.tables[table]
        # Dispatch to an appropriate implementation.
        if self._connector.driver == "mysql":
            self._upsert_onduplicatekey(table, items)
        else:
            self._upsert_generic(table, items)

    def _upsert_generic(self, table, items):
        """Upsert a batch of items one at a time, trying INSERT then UPDATE.

        This is a tremendously inefficient way to write a batch of items,
        but it's guaranteed to work without special cooperation from the
        database.  For MySQL we use the much improved _upsert_onduplicatekey.
        """
        userid = items[0].get("userid")
        for item in items:
            assert item.get("userid") == userid
            try:
                # Try to insert the item.
                # If it already exists, this fails with an integrity error.
                self.execute(insert(table), item)
            except IntegrityError:
                # Update the item.
                # Use the table's primary key fields in the WHERE clause,
                # and put all other fields into the UPDATE clause.
                item = item.copy()
                query = update(table)
                for key in table.primary_key:
                    try:
                        query = query.where(key == item.pop(key.name))
                    except KeyError:
                        msg = "Item is missing primary key column %r"
                        raise ValueError(msg % (key.name,))
                query = query.values(**item)
                self.execute(query, item)

    def _upsert_onduplicatekey(self, table, items):
        """Upsert a batch of items using the ON DUPLICATE KEY UPDATE syntax.

        This is a custom batch upsert implementation based on non-standard
        features of MySQL.  The resulting query will be something like the
        following, where M is the number of fields in each item and N is the
        number of items being inserted:

            INSERT INTO table (c1, ..., cM)
            VALUES (:c11, ..., :cM1), ..., (:c1N, ... :cMN)
            ON DUPLICATE KEY UPDATE c1 = VALUES(c1), ..., cM = VALUES(cM)

        The values from the given items will be collected into a matching set
        of bind parameters :c11 through :cMN  when executing the query.
        """
        userid = items[0].get("userid")
        # Group the items to be inserted into batches that all have the same
        # set of fields.  Each batch will have the same ON DUPLICATE KEY UPDATE
        # clause and so can be sent as a single query.
        batches = defaultdict(list)
        for item in items:
            assert item.get("userid") == userid
            batches[frozenset(item.iterkeys())].append(item)
        # Now construct and send an appropriate query for each batch.
        for batch in batches.itervalues():
            # Since we're crafting SQL by hand, assert that each field is
            # actually a plain alphanum field name.  Can't be too careful...
            fields = batch[0].keys()
            assert all(field.isalnum() for field in fields)
            # Each item corresponds to a set of bindparams and a matching
            # entry in the "VALUES" clause of the query.
            query = "INSERT INTO %s (%s) VALUES "\
                    % (table.name, "".join(fields))
            binds = [":%s%%(num)d" % field for field in fields]
            pattern = "(%s) " % ",".join(binds)
            params = {}
            vclauses = []
            for num, item in enumerate(batch):
                vclauses.append(pattern % {"num": num})
                for field, value in item.iteritems():
                    params["%s%d" % (field, num)] = value
            query += ",".join(vclauses)
            # The ON DUPLICATE KEY CLAUSE updates all the given fields.
            updates = ["%s = VALUES(%s)" % (field, field) for field in fields]
            query += " ON DUPLICATE KEY UPDATE" + ",".join(updates)
            # Now we can execute it as one big query.
            self.execute(sqltext(query), params)
